#!/usr/bin/env python     

__author__ = "Marta Bañón (mbanon)"
__version__ = "Version 0.1 # 21/06/2019 # Initial release # Marta Bañón"
__version__ = "Version 0.2 # 23/07/2019 # Non-Redis Bifixer # Marta Bañón"
__version__ = "Version 0.3 # 20/08/2019 # New feature: Segmentation # Marta Bañón"
__version__ = "Version 0.4 # 15/06/2021 # Easy install # Elsa Sarrías"
__version__ = "Version 0.5 # 22/06/2021 # Replacements improvements and fix tokenization for Maltese # Jaume Zaragoza"
__version__ = "Version 0.7 # 15/02/2022 # Disable punctuation normalization # Jaume Zaragoza"
__version__ = "Version 0.8 # 15/06/2022 # Remove HTML tags # Marta Bañón"

import os
import sys
import argparse
import copy
import traceback
import logging

from unicodedata import category as cat
from unidecode import unidecode
from tempfile import gettempdir
from timeit import default_timer
from xxhash import xxh64

try:
    from . import util
    from . import restorative_cleaning
    from . import segmenter
except (ImportError, SystemError):
    import util
    import restorative_cleaning
    import segmenter

# Translate table to remove non alphabetic characters
tbl = [chr(i) for i in range(sys.maxunicode) if not cat(chr(i)).startswith('L')]
remove_non_alpha = str.maketrans('', '', ''.join(tbl))

def initialization():
    global ilines
    global olines

    ilines = 0
    olines = 0
    header = "--header" in sys.argv

    logging.info("Processing arguments...")
    parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]), formatter_class=argparse.ArgumentDefaultsHelpFormatter, description=__doc__)

    # Mandatory parameters
    # Input file
    parser.add_argument('input', type=argparse.FileType('rt', encoding='UTF-8'), default=None, help="Tab-separated files to be bifixed")
    # Output file (corpus)
    parser.add_argument('output', type=argparse.FileType('w'), default=sys.stdout, help="Fixed corpus")
    # Source language
    parser.add_argument("srclang", type=str, help="Source language (SL) of the input")
    # Target language
    parser.add_argument("trglang", type=str, help="Target language (TL) of the input")

    # Options group
    groupO = parser.add_argument_group('Optional')
    # Format
    groupO.add_argument("--header", action='store_true', help="Input file will have header")
    groupO.add_argument("--scol", type=util.check_positive if not header else str, default=3 if not header else "src_text", help="Source sentence column (starting in 1). The name of the field is expected instead of the position if --header is set")
    groupO.add_argument("--tcol", type=util.check_positive if not header else str, default=4 if not header else "trg_text", help="Target sentence column (starting in 1). The name of the field is expected instead of the position if --header is set")
    groupO.add_argument("--sdeferredcol", type=util.check_positive if not header else str, help="Source deferred standoff annotation column (starting in 1). The name of the field is expected instead of the position if --header is set")
    groupO.add_argument("--tdeferredcol", type=util.check_positive if not header else str, help="Target deferred standoff annotation column (starting in 1). The name of the field is expected instead of the position if --header is set")
    groupO.add_argument("--sparagraphid", type=util.check_positive if not header else str, help="Source paragraph identification column (starting in 1). The name of the field is expected instead of the position if --header is set")
    groupO.add_argument("--tparagraphid", type=util.check_positive if not header else str, help="Target paragraph identification column (starting in 1). The name of the field is expected instead of the position if --header is set")

    # Character fixing
    groupO.add_argument('--ignore_characters', default=False, action='store_true', help="Doesn't fix mojibake or other character issues")
    groupO.add_argument('--ignore_normalization', default=False, action='store_true', help="Doesn't normalize punctuation and spaces.")

    # HTML tags
    groupO.add_argument('--ignore_html', default=False, action='store_true', help="Doesn't remove HTML tags")

    # Empty sides
    groupO.add_argument('--ignore_empty', default=False, action='store_true', help="Doesn't remove sentences with empty source or target")
    
    # Too long sides
    groupO.add_argument('--ignore_long', default=False, action='store_true', help="Doesn't ignore too long sentences")
    
    # Orthography
    groupO.add_argument('--ignore_orthography', default=False, action='store_true', help="Doesn't apply orthography fixing")

    # Common detokenization issues
    groupO.add_argument('--ignore_detokenization', default=False, action='store_true', help="Doesn't fix common tokenization issues")

    # Deduplication
    groupO.add_argument('--ignore_duplicates', default=False, action='store_true', help="Doesn't obtain the hashes of parallel sentences")
    groupO.add_argument('--aggressive_dedup', default=False, action='store_true', help="Treats similar sentences as duplicates (marking them with the same hash)")

    # Segmentation
    groupO.add_argument('--ignore_segmentation', default=False, action='store_true', help="Doesn't change segmentation of long sentences")
    groupO.add_argument('--words_before_segmenting', default=15, type=util.check_positive, help="Max words allowed in one side of a parallel sentence before trying to segmentate it. Set to 0 to applicate segmentation on everything.")
    groupO.add_argument('--segmenter', default="nltk", type=str, choices=["nltk", "loomchild"], help="Segmenter module.")
    groupO.add_argument('--tmp_dir', default=gettempdir(), help="Temporary directory where creating the temporary files of this program")

    # Logging group
    groupL = parser.add_argument_group('Logging')
    groupL.add_argument('-q', '--quiet', action='store_true', help='Silent logging mode')
    groupL.add_argument('--debug', action='store_true', help='Debug logging mode')
    groupL.add_argument('--logfile', type=argparse.FileType('a'), default=sys.stderr, help="Store log to a file")
    groupL.add_argument('-v', '--version', action='version', version="%(prog)s " + __version__, help="show version of this script and exit")

    # Validating & parsing
    args = parser.parse_args()
    util.logging_setup(args)
    args.dedup = not args.ignore_duplicates  # more friendly usage of the ignore_duplicates flag

    logging.debug("Arguments processed: {}".format(str(args)))

    logging.info("Arguments processed.")

    return args


def fix_sentences(args):
    if ('ilines' in globals() and 'olines' in globals()):
        global ilines
        global olines
    else:
        ilines = 0
        olines = 0

    if not args.ignore_characters:
        chars_slang, charsRe_slang = restorative_cleaning.getCharsReplacements(args.srclang)
        chars_tlang, charsRe_tlang = restorative_cleaning.getCharsReplacements(args.trglang)

    if not args.ignore_normalization:
        punctChars_slang, punctRe_slang = restorative_cleaning.getNormalizedPunctReplacements(args.srclang)
        punctChars_tlang, punctRe_tlang = restorative_cleaning.getNormalizedPunctReplacements(args.trglang)
        
    if not args.ignore_orthography:
        replacements_slang = restorative_cleaning.getReplacements(args.srclang)
        replacements_tlang = restorative_cleaning.getReplacements(args.trglang)

    if not args.ignore_detokenization:
        detoks_slang = restorative_cleaning.getDetokenizations(args.srclang)
        detoks_tlang = restorative_cleaning.getDetokenizations(args.trglang)

    if not args.ignore_segmentation:
        source_segmenter = segmenter.NaiveSegmenter(args.srclang, args.segmenter)
        target_segmenter = segmenter.NaiveSegmenter(args.trglang, args.segmenter)

    if args.header:
        header = next(args.input).strip().split("\t")

        # Transform fields to idxs
        if args.scol not in header:
            raise Exception(f"The provided --scol '{args.scol}' is not in the input header")
        if args.tcol not in header:
            raise Exception(f"The provided --tcol '{args.tcol}' is not in the input header")

        args.scol = int(header.index(args.scol)) + 1
        args.tcol = int(header.index(args.tcol)) + 1

        if args.sdeferredcol:
            if args.sdeferredcol not in header:
                raise Exception(f"The provided --sdeferredcol '{args.sdeferredcol}' is not in the input header")

            args.sdeferredcol = int(header.index(args.sdeferredcol)) + 1
        if args.tdeferredcol:
            if args.tdeferredcol not in header:
                raise Exception(f"The provided --tdeferredcol '{args.tdeferredcol}' is not in the input header")

            args.tdeferredcol = int(header.index(args.tdeferredcol)) + 1
        if args.sparagraphid:
            if args.sparagraphid not in header:
                raise Exception(f"The provided --sparagraphid '{args.sparagraphid}' is not in the input header")

            args.sparagraphid = int(header.index(args.sparagraphid)) + 1
        if args.tparagraphid:
            if args.tparagraphid not in header:
                raise Exception(f"The provided --tparagraphid '{args.tparagraphid}' is not in the input header")

            args.tparagraphid = int(header.index(args.tparagraphid)) + 1

        # Write the output header once
        args.output.write("\t".join(header))

        if args.dedup:
            args.output.write("\tbifixer_hash\tbifixer_score")

        args.output.write("\n")

    for i in args.input:
        ilines += 1
        parts = i.split("\t")

        try:
            source_sentence = parts[args.scol - 1]
            target_sentence = parts[args.tcol - 1]

            # Check optional indexes
            if args.sdeferredcol and args.tdeferredcol:
                parts[args.sdeferredcol - 1]
                parts[args.tdeferredcol - 1]
            if args.sparagraphid and args.tparagraphid:
                parts[args.sparagraphid - 1]
                parts[args.tparagraphid - 1]
        except IndexError:
            logging.error(traceback.format_exc())
            logging.error("Wrong column index on line " + str(ilines))
            continue

        very_long = False

        # None of source or target sentences is empty:
        if args.ignore_empty or (source_sentence and target_sentence):
            if not args.ignore_long and (len(source_sentence) > 5000 or len(target_sentence) > 5000):
                very_long = True

            if not args.ignore_characters and not very_long:
                fixed_source = restorative_cleaning.fix(source_sentence, args.srclang, chars_slang, charsRe_slang)
                fixed_target = restorative_cleaning.fix(target_sentence, args.trglang, chars_tlang, charsRe_tlang)
            else:
                fixed_source = source_sentence.strip(" \n") 
                fixed_target = target_sentence.strip(" \n") 

            if not args.ignore_html and not very_long:
                fixed_source = restorative_cleaning.remove_html_tags(fixed_source)
                fixed_target = restorative_cleaning.remove_html_tags(fixed_target)            
                
            if not args.ignore_normalization and not very_long:
                fixed_source = restorative_cleaning.normalize(fixed_source, args.srclang, punctChars_slang, punctRe_slang)
                fixed_target = restorative_cleaning.normalize(fixed_target, args.trglang, punctChars_tlang, punctRe_tlang)

            if not args.ignore_orthography and not very_long:
                corrected_source = restorative_cleaning.ortho_detok_fix(fixed_source, replacements_slang, detoks_slang)
                corrected_target = restorative_cleaning.ortho_detok_fix(fixed_target, replacements_tlang, detoks_tlang)
            else:
                corrected_source = fixed_source
                corrected_target = fixed_target

            if not args.ignore_segmentation and (len(fixed_source.split()) > args.words_before_segmenting or len(fixed_target.split()) > args.words_before_segmenting) and not very_long:
                # The naive_segmenter must return an array of tuples (source sentence, target sentence)
                segments = segmenter.naive_segmenter(source_segmenter, target_segmenter, corrected_source, corrected_target)
            else:
                # keep original segmentation
                segments = [{"source_segment": corrected_source, "target_segment": corrected_target}]

            sent_num = 0
            for segment in segments:
                if not args.ignore_empty and (len(segment["source_segment"]) == 0 or len(segment["target_segment"]) == 0):
                    continue
                if args.dedup:
                    if args.aggressive_dedup:
                        normalized_src = unidecode(segment["source_segment"].lower().translate(remove_non_alpha))
                        normalized_trg = unidecode(segment["target_segment"].lower().translate(remove_non_alpha))

                        segment_hash = xxh64(normalized_src + "\t" + normalized_trg).hexdigest()

                        charsum = sum(ord(ch) for ch in segment["source_segment"] + segment["target_segment"])

                        ranking = round(charsum / len(segment["source_segment"] + segment["target_segment"]), 2)

                    else:
                        segment_hash = xxh64(segment["source_segment"] + "\t" + segment["target_segment"]).hexdigest()
                        ranking = 1
                # if  dedupping: Add extra columnsn with hash and ranking in output file
                # Restored parts object, with the fixed segment, overwritten for each pair of extra segments,
                new_parts = copy.deepcopy(parts)

                new_parts[args.scol - 1] = segment["source_segment"]
                new_parts[args.tcol - 1] = segment["target_segment"]

                if len(segments) > 1:
                    sent_num += 1

                    if args.sdeferredcol and args.tdeferredcol:
                        if "#" in parts[args.sdeferredcol - 1] or "#" in parts[args.tdeferredcol - 1]:
                            # Reconstruction
                            if "#" in parts[args.sdeferredcol - 1]:
                                if sent_num != int(parts[args.sdeferredcol - 1].split('#')[1]):
                                    continue
                            elif "#" in parts[args.tdeferredcol - 1]:
                                if sent_num != int(parts[args.tdeferredcol - 1].split('#')[1]):
                                    continue
                        else:
                            new_parts[args.sdeferredcol - 1] = parts[args.sdeferredcol - 1].rstrip("\n") + "#" + str(sent_num)
                            new_parts[args.tdeferredcol - 1] = parts[args.tdeferredcol - 1].rstrip("\n") + "#" + str(sent_num)
                    if args.sparagraphid and args.tparagraphid:
                        new_parts[args.sparagraphid - 1] = parts[args.sparagraphid - 1].rstrip("\n") + "#" + str(sent_num)
                        new_parts[args.tparagraphid - 1] = parts[args.tparagraphid - 1].rstrip("\n") + "#" + str(sent_num)

                # sentence sides may be empty now because it contained only spaces or similar weird thing
                # for sentences containing only spaces but not normalized, strip them
                if args.ignore_empty or (new_parts[args.scol - 1].strip() and new_parts[args.tcol - 1].strip()):
                    if (args.dedup):
                        # Remove the "/n" at the end of the last item
                        new_parts[-1] = str(new_parts[-1]).strip("\n")

                        new_parts.append(segment_hash)  # hash and ranking are added at the end
                        new_parts.append(ranking)
                        args.output.write("\t".join(str(v) for v in new_parts) + "\n")  # Convert to strings
                        # Remove hash and ranking for next iterations of loop
                        new_parts.pop()
                        new_parts.pop()
                    else:
                        # When no deduplicating:
                        new_parts[-1] = str(new_parts[-1]).strip("\n")
                        args.output.write("\t".join(str(v) for v in new_parts) + "\n")
                    olines += 1
                else:
                    # empty sides after processing
                    continue

        else:
            # source and/or target is empty
            continue


def perform_fixing(args):
    global ilines
    global olines

    time_start = default_timer()
    logging.info("Starting fixing text")
    fix_sentences(args)
    logging.info("Text fixing finished")

    # Stats
    logging.info("Finished")
    elapsed_time = default_timer() - time_start
    logging.info("Input lines: {0} rows".format(ilines))
    logging.info("Output lines: {0} rows".format(olines))
    logging.info("Elapsed time {0:.2f} s".format(elapsed_time))
    logging.info("Troughput: {0} rows/s".format(int((ilines * 1.0) / elapsed_time)))

    logging.info("Output file: {0}".format(os.path.abspath(args.output.name)))


def main(args):
    logging.info("Executing main program...")
    perform_fixing(args)
    logging.info("Program finished")


if __name__ == '__main__':
    try:
        util.logging_setup()
        args = initialization()  # Parsing parameters
        main(args)  # Running main program
    except Exception as ex:
        tb = traceback.format_exc()
        logging.error(tb)
        sys.exit(1)



#!/usr/bin/env python
import logging
import ftfy
import os
import regex
import re
import html

global_chars_lang = {}

chars3Re = regex.compile("[\uE000-\uFFFF]")
chars3Re2 = regex.compile("[\u2000-\u200F]")
chars3Re3 = regex.compile("\u007F|[\u0080-\u00A0]")
quotesRegex = regex.compile("(?P<start>[[:alpha:]])\'\'(?P<end>(s|S|t|T|m|M|d|D|re|RE|ll|LL|ve|VE|em|EM)\W)")
collapse_spaced_entities = regex.compile('([&][ ]*[#][ ]*)([0-9]{2,6})([ ]*[;])')
html_tags_regex = re.compile('<.*?>') 

#https://en.wikipedia.org/wiki/CJK_Symbols_and_Punctuation
cjk_langs = [
    "ja", #japanese
    "ko", #korean
    "zh", #chinese
    "zh-hant" #traditional chinese
]

def getCharsReplacements(lang):
    # languages that use cyrillic alphabet
    # Check https://www.tug.org/TUGboat/tb17-2/tb51pisk.pdf and/or
    # https://www.quora.com/Which-languages-are-written-in-Cyrillic-script
    # for a comprehensive list
    # Some of the langs does accept both cyrillic and latin, or is migrating to latin
    cyrillic_langs = [
        "ab",  # abkhazian
        "av",  # avar/avaric
        "az",  # azerbaijani
        "ba",  # bashkir
        "be",  # belarusian
        "bg",  # bulgarian
        "bs",  # bosnian
        "ce",  # chechen
        "cv",  # chuvash
        "kk",  # kazakh
        "ku",  # kurdish
        "kv",  # komi
        "ky",  # kirghiz/kyrgyz
        "mk",  # macedonian
        "mn",  # mongolian
        "os",  # ossetic/ossetian
        "ru",  # russian
        "sr",  # serbian
        "tg",  # tajik/tadzhik
        "tk",  # turkmen
        "tt",  # tatar
        "ug",  # uighur
        "uk",  # ukranian
        "uz"  # uzbek
    ]

    # https://en.wikipedia.org/wiki/Caron
    # http://diacritics.typo.cz/index.php?id=5
    langs_with_carons = [
        "cs",  # czech
        "et",  # estonian
        "fi",  # finnish
        "hr",  # croatian
        "ln",  # lingala
        "lt",  # lithuanian
        "lv",  # latvian
        "sk",  # slovak
        "sl",  # slovenian
        "sr",  # serbian
        "yo"  # yoruba
    ]
    


    # Annoying characters, common for all languages
    chars = {
        '\u2028': ' ',  # line separators (\n)
        '&#10;': " ",  # \n
        '&#9;': " ",  # \t
        '&#10': " ",  # \n
        '&#9': " ",  # \t
        '\n': "",
        '\u000C' : " ", # \v vertical tab
        '\u000D' : " ", # \f form feed
        '&#xa': "",
        '&#xA': "",

        '\u000D': "",  # carriage returns (\r)
        '&#13;': " ",
        '&#xd;': " ",
        '&#xD;': " ",

        # unicode ligatures
        '\uFB00': 'ff',
        '\uFB01': 'fi',
        '\uFB02': 'fl',
        '\uFB03': 'ffi',
        '\uFB04': 'ffl',
        '\uFB06': 'st',

        '&nbsp;': " ",
        '&lt;': "<",
        '&gt;': ">",
        '&amp;': "&",
        '&quot;': '"',
        '&apos;': "'",
        '&iexcl;': '¡',
        '&cent;': '¢',
        '&pound;': '£',

        '&Agrave;': 'À',
        '&Aacute;': 'Á',
        '&Acirc;': 'Â',
        '&Atilde;': 'Ã',
        '&Auml;': 'Ä',
        '&Aring;': 'Å',
        '&Aelig;': 'Æ',

        '&agrave;': 'à',
        '&aacute;': 'á',
        '&acirc;': 'â',
        '&atilde;': 'ã',
        '&auml;': 'ä',
        '&aring;': 'å',
        '&aelig;': 'æ',

        '&Ccedil;': 'Ç',
        '&ccedil;': 'ç',

        '&Egrave;': 'È',
        '&Eacute;': 'É',
        '&Ecirc;': 'Ê',
        '&Euml;': 'Ë',

        '&egrave;': 'è',
        '&eacute;': 'é',
        '&ecirc;': 'ê',
        '&euml;': 'ë',
        '&Igrave;': 'Ì',
        '&Iacute;': 'Í',
        '&Icirc;': 'Î',
        '&Iuml;': 'Ï',

        '&igrave;': 'ì',
        '&iacute;': 'í',
        '&icirc;': 'î',
        '&iuml;': 'ï',

        '&Ntilde;': 'Ñ',
        '&ntilde;': 'ñ',

        '&Ograve;': 'Ò',
        '&Oacute;': 'Ó',
        '&Ocirc;': 'Ô',
        '&Otilde;': 'Õ',
        '&Ouml;': 'Ö',

        '&ograve;': 'ò',
        '&oacute;': 'ó',
        '&ocirc;': 'ô',
        '&otilde;': 'õ',
        '&ouml;': 'ö',

        '&times;': '×',  # ×
        '&Oslash;': '',  # Ø
        '&oslash;': '',  # ø

        '&Ugrave;': 'Ù',
        '&Uacute;': 'Ú',
        '&Ucirc;': 'Û',
        '&Uuml;': 'Ü',

        '&ugrave;': 'ù',
        '&uacute;': 'ú',
        '&ucirc;': 'û',
        '&uuml;': 'ü',

        '&Yacute;': 'Ý',
        '&yacute;': 'ý',
        '&yuml;': 'ÿ',  # ÿ
        '&Yuml;': 'Ÿ',  # capital Y with diaeres Ÿ

        '&thorn;': 'Þ',  # þ
        '&szlig;': 'ß',  # ß

        '&divide;': '÷',  # ÷
        '&euro;': '€',

        '\u02C1': "¿",  # ˁ -> ?
        '\u02C2': "<",  # ˂ -> <
        '\u02C3': ">",  # ˃ -> >

        # https://www.utf8-chartable.de/unicode-utf8-table.pl?number=1024&names=2&utf8=char

        'Â¿': '¿',
        'Â¡': '¡',

        'Ã€': 'À',
        'Ã<80>': 'À',
        'Ã<81>': 'Á',
        'Ã<82>': 'Â',
        'Ã<83>': 'Ã',
        'Ã<84>': 'Ä',
        'Ã<85>': 'Å',
        'Ã‚': 'Â',
        'Ãƒ': 'Ã',
        'Ã„': 'Ä',
        'Ã…': 'Å',
        'Ã<86>': 'Æ',
        'Ã†': 'Æ',
        'Ä<80>': 'Ā',
        'Ä<82>': 'Ă',

        'Ã¡': 'á',
        'Ã¤': 'ä',
        'Ã¢': 'â',
        'Ã£': 'ã',
        'Ã¥': 'å',
        'Ã¦': 'æ',
        'Ä<81>': 'ā',
        'Ä<83>': 'ă',

        'Ã<87>': 'Ç',
        'Ã‡': 'Ç',
        'Ã§': 'ç',
        'ÄŤ': 'č',
        'ДЌ': 'č',
        'Д‡': 'ć',
        "ÄŚ": "Č",

        'Ã<88>': 'È',
        'Ã<89>': 'É',
        'Ã<8A>': 'Ê',
        'Ã<8B>': 'Ë',
        'Ãˆ': 'È',
        'Ã‰': 'É',
        'ÃŠ': 'Ê',
        'Ã‹': 'Ë',

        'Ã¨': 'è',
        'Ã©': 'é',
        'Ãª': 'ê',
        'Ã«': 'ë',

        'Ã<8C>': 'Ì',
        'Ã<8E>': 'Î',
        'ÃŒ': 'Ì',
        #'Ã<8D>': 'Í',
        'ÃŽ': 'Î',
        'Ã<8F>': 'Ï',
        'Ã¬': 'ì',
        'Ã<AD>': 'í',
        'Ã®': 'î',
        'Ã¯': 'ï',

        'Ä¾': 'ľ',

        'Ã’': 'Ò',
        'Ã<92>': 'Ò',
        'Ã<93>': 'Ó',
        'Ã<94>': 'Ô',
        'Ã<95>': 'Õ',
        'Ã<96>': 'Ö',
        'Ã“': 'Ó',
        'Ã”': 'Ô',
        'Ã•': 'Õ',
        'Ã–': 'Ö',
        'Ã˜': 'Ø',
        'Ã<98>': 'Ø',
        'Å’': 'Œ',

        'Ã²': 'ò',
        'Ã³': 'ó',
        'Ã´': 'ô',
        'Ãµ': 'õ',
        'Ã¸': 'ø',
        'Å“': 'œ',
        'Ã¶': 'ö',

        'Ã‘': 'Ñ',
        'Ã<91>': 'Ñ',
        'Ã±': 'ñ',

        'Å¡': 'š',
        'ĹĄ': 'š',
        'ЕЎ': 'š',
        "Ĺˇ": "š",
        "Å¡": "š",
        "Ĺ ": "Š",

        'Å¥': 'ť',

        'Ã<99>': 'Ù',
        'Ã<9A>': 'Ú',
        'Ã<9B>': 'Û',
        'Ã<9C>': 'Ü',
        'Ã™': 'Ù',
        'Ãš': 'Ú',
        'Ã›': 'Û',
        'Ãœ': 'Ü',

        'Ã¹': 'ù',
        'Ãº': 'ú',
        'Ã¼': 'ü',
        'Ã»': 'û',

        'Ã<9D>': 'Ý',
        'Å¸': 'Ÿ',

        'Ã½': 'ý',
        'Ã¿': 'ÿ',
        'Å½': 'Ž',
        'Å¾': 'ž',
        'Ĺľ': 'ž',
        'Ĺˇ': 'ž',
        'Ĺž': 'ž',
        "Ĺ˝": "Ž",
        "Å3⁄4": "ž",

        'Ãž': 'Þ',
        'Ã<9E>': 'Þ',
        'Ã¾': 'þ',

        'Ã<90>': 'Ð',
        'ÃŸ': 'ß',
        'Ã<9F>': 'ß',
        'Âµ': 'µ',
        'Ã°': 'ð',

    }

    if lang.lower() not in langs_with_carons:
        chars['\u0165'] = "t'"  # latin small letter t with caron ť
        chars['\u0192'] = "f"  # latin small letter f with hook ƒ

    if lang.lower() not in cyrillic_langs:
        # Cyrilic charcaters replaced to latin characters

        chars['Ѐ'] = 'È'
        chars['Ё'] = 'Ë'
        chars['Ѕ'] = 'S'
        chars['Ї'] = 'Ï'
        chars['І'] = 'I'
        chars['Ј'] = 'J'
        chars['А'] = 'A'
        chars['В'] = 'B'
        chars['Е'] = 'E'
        chars['К'] = 'K'
        chars['М'] = 'M'
        chars['Н'] = 'H'
        chars['О'] = 'O'
        chars['Р'] = 'P'
        chars['С'] = 'C'
        chars['Т'] = 'T'
        chars['У'] = 'y'
        chars['Х'] = 'X'
        chars['Ь'] = 'b'
        chars['ѐ'] = 'è'
        chars['ё'] = 'ë'
        chars['а'] = 'a'
        chars['в'] = 'B'
        chars['г'] = 'r'
        chars['е'] = 'e'
        chars['к'] = 'k'
        chars['м'] = 'M'
        chars['н'] = 'H'
        chars['о'] = 'o'
        chars['р'] = 'p'
        chars['с'] = 'c'
        chars['т'] = 'T'
        chars['у'] = 'y'
        chars['х'] = 'x'
        chars['ь'] = 'b'
        chars['ѕ'] = 's'
        chars['і'] = 'i'
        chars['ї'] = 'ï'
        chars['ј'] = 'j'
        chars['Ү'] = 'Y'
        chars['Ү'] = 'Y'
        chars['Һ'] = 'h'
        chars['һ'] = 'h'
        chars['Ӏ'] = 'I'
        chars['ӏ'] = 'I'
        chars['Ӓ'] = 'Ä'
        chars['ӓ'] = 'ä'
        chars['Ԁ'] = 'd'
        chars['ԁ'] = 'd'
        chars['Ԛ'] = 'Q'
        chars['ԛ'] = 'q'
        chars['Ԝ'] = 'W'
        chars['ԝ'] = 'w'
        chars['Ꙇ'] = 'l'
        chars['ꙇ'] = 'l'
        chars['Ꚃ'] = 'S'
        chars['ꚃ'] = 'S'
        chars['\u0443'] = 'y'  #

    # Homoglyph Greek replacements for non-greek languages
    if lang.lower() != 'el':
        chars['Α'] = 'A'
        chars['α'] = 'a'
        chars['Β'] = 'B'
        if lang.lower() != 'de':
            chars['β'] = 'B'
        else:
            chars['β'] = 'ß'
        chars['γ'] = 'Y'
        chars['Δ'] = 'A'
        chars['Ε'] = 'E'
        chars['ε'] = 'e'
        chars['Ζ'] = 'Z'
        chars['ζ'] = 'Z'
        chars['Η'] = 'H'
        chars['η'] = 'n'
        chars['Ι'] = 'I'
        chars['Κ'] = 'K'
        chars['κ'] = 'K'
        chars['Λ'] = 'A'
        chars['Μ'] = 'M'
        chars['μ'] = 'u'
        chars['Ν'] = 'N'
        chars['ν'] = 'v'
        chars['Ξ'] = 'E'
        chars['ξ'] = 'E'
        chars['Ο'] = 'O'
        chars['ο'] = 'o'
        chars['Ρ'] = 'P'
        chars['ρ'] = 'p'
        chars['Σ'] = 'E'
        chars['Τ'] = 'T'
        chars['τ'] = 't'
        chars['Υ'] = 'Y'
        chars['υ'] = 'u'
        chars['Χ'] = 'X'
        chars['χ'] = 'x'
        chars['ω'] = 'w'

    if lang.lower() == "de":
        # Remove and/or replace certain keys from 'chars' in German
        chars['&bdquo;'] = '„'
        chars['\u201E'] = '„'
        chars['\u00D8'] = 'Ø'  # latin capital letter o with stroke Ø
        chars['\u00F8'] = 'ø'  # latin small letter o with stroke ø
        chars['&Oslash;'] = 'Ø'  # Ø
        chars['&oslash;'] = 'ø'
    else:
        chars['׳'] = "'"
        chars['״'] = '"'
        chars['־'] = '-'
        chars['׀'] = '|'
        chars['׃'] = ':'

    # Normalize single quote sometimes used as apostrophe in Maltese
    if lang.lower() == 'mt':
        chars['’'] = "'"

    if lang.lower() != "el":
        # Greek Letters
        chars['&Alpha;'] = 'A'  # Alpha   Α -> Changed to latin A
        chars['Α'] = 'A'  # Alpha  Α -> Changed to latin A
        chars['&Beta;'] = 'B'  # Beta     Β -> Changed to latin B
        chars['Β'] = 'B'  # Beta   Β -> Changed to latin B
        chars['&Gamma;'] = 'Γ'  # Gamma   Γ
        chars['&Delta;'] = 'Δ'  # Delta   Δ
        chars['&Epsilon;'] = 'E'  # Epsilon       Ε -> Changed to latin E
        chars['Ε'] = 'E'  # Epsilon        Ε -> Changed to latin E
        chars['&Zeta;'] = 'Z'  # Zeta     Ζ -> Changed to latin Z
        chars['Ζ'] = 'Z'  # Zeta   Ζ -> Changed to latin Z
        chars['&Eta;'] = 'H'  # Eta       Η -> Changed to latin H
        chars['Η'] = 'H'  # Eta    Η -> Changed to latin H
        chars['&Theta;'] = 'Θ'  # Theta   Θ
        chars['&Iota;'] = 'I'  # Iota     Ι -> Chaged to latin I
        chars['Ι'] = 'I'  # Iota   Ι -> Chaged to latin I
        chars['&Kappa;'] = 'K'  # Kappa   Κ -> Changed to latin K
        chars['Κ'] = 'K'  # Kappa  Κ -> Changed to latin K
        chars['&Lambda;'] = 'Λ'  # Lambda Λ
        chars['&Mu;'] = 'M'  # Mu Μ -> Changed to latin M
        chars['Μ'] = 'M'  # Mu     Μ -> Changed to latin M
        chars['&Nu;'] = 'N'  # Nu Ν -> Changed to latin N
        chars['Ν'] = 'N'  # Nu     Ν -> Changed to latin N
        chars['&Xi;'] = 'Ξ'  # Xi Ξ
        chars['&Omicron;'] = 'O'  # Omicron       Ο -> Changed to latin O
        chars['Ο'] = 'O'  # Omicron        Ο -> Changed to latin O
        chars['&Pi;'] = 'Π'  # Pi Π
        chars['&Rho;'] = 'P'  # Rho       Ρ -> Changed to latin P
        chars['Ρ'] = 'P'  # Rho    Ρ -> Changed to latin P
        chars['&Sigma;'] = 'Σ'  # Sigma   Σ
        chars['&Tau;'] = 'T'  # Tau       Τ -> Changed to latin T
        chars['Τ'] = 'T'  # Tau    Τ -> Changed to latin T
        chars['&Upsilon;'] = 'Y'  # Upsilon       Υ -> Changed to latin Y
        chars['Υ'] = 'Y'  # Upsilon        Υ -> Changed to latin Y
        chars['&Phi;'] = 'Φ'  # Phi       Φ
        chars['&Chi;'] = 'X'  # Chi       Χ -> Changed to latin X
        chars['Χ'] = 'X'  # Chi    Χ -> Changed to latin X
        chars['&Psi;'] = 'Ψ'  # Psi       Ψ
        chars['&Omega;'] = 'Ω'  # Omega   Ω
        chars['&alpha;'] = 'a'  # alpha   α -> Changed to latin a
        chars['α'] = 'a'  # alpha  α -> Changed to latin a
        chars['&beta;'] = 'β'  # beta     β
        chars['&gamma;'] = 'γ'  # gamma   γ
        chars['&delta;'] = 'δ'  # delta   δ
        chars['&epsilon;'] = 'ε'  # epsilon       ε
        chars['&zeta;'] = 'ζ'  # zeta     ζ
        chars['&eta;'] = 'n'  # eta       η -> Changed to latin n
        chars['η'] = 'n'  # eta    η -> Changed to latin n
        chars['&theta;'] = 'θ'  # theta   θ
        chars['&iota;'] = 'ι'  # iota     ι
        chars['&kappa;'] = 'k'  # kappa   κ -> Changed to latin k
        chars['κ'] = 'k'  # kappa  κ -> Changed to latin k
        chars['&lambda;'] = 'λ'  # lambda λ
        chars['&mu;'] = 'μ'  # mu μ
        chars['&nu;'] = 'v'  # nu ν -> Changed to latin v
        chars['ν'] = 'v'  # nu     ν -> Changed to latin v
        chars['&xi;'] = 'ξ'  # xi ξ
        chars['&omicron;'] = 'o'  # omicron       ο -> Changed to latin o
        chars['ο'] = 'o'  # omicron        ο -> Changed to latin o
        chars['&pi;'] = 'π'  # pi π
        chars['&rho;'] = 'p'  # rho       ρ -> Changed to latin p
        chars['ρ'] = 'p'  # rho    ρ -> Changed to latin p
        chars['&sigmaf;'] = 'ς'  # sigmaf ς
        chars['&sigma;'] = 'σ'  # sigma   σ
        chars['&tau;'] = 't'  # tau       τ -> Changed to latin t
        chars['τ'] = 't'  # tau    τ -> Changed to latin t
        chars['&upsilon;'] = 'u'  # upsilon       υ -> Changed to latin u
        chars['υ'] = 'u'  # upsilon        υ -> Changed to latin u
        chars['&phi;'] = 'φ'  # phi       φ
        chars['&chi;'] = 'χ'  # chi       χ
        chars['&psi;'] = 'ψ'  # psi       ψ
        chars['&omega;'] = 'ω'  # omega   ω
        chars['&thetasym;'] = 'ϑ'  # theta symbol ϑ
        chars['&upsih;'] = 'ϒ'  # upsilon symbol  ϒ
        chars['&piv;'] = 'ϖ'  # pi symbol ϖ
        chars['\u03BF'] = 'o'  # GREEK SMALL LETTER OMICRON
    else:
        # Greek Letters
        chars['&Alpha;'] = 'Α'  # Alpha
        chars['&Beta;'] = 'Β'  # Beta
        chars['&Gamma;'] = 'Γ'  # Gamma   Γ
        chars['&Delta;'] = 'Δ'  # Delta   Δ
        chars['&Epsilon;'] = 'Ε'  # Epsilon
        chars['&Zeta;'] = 'Ζ'  # Zeta
        chars['&Eta;'] = 'Η'  # Eta
        chars['&Theta;'] = 'Θ'  # Theta   Θ
        chars['&Iota;'] = 'Ι'  # Iota
        chars['&Kappa;'] = 'Κ'  # Kappa
        chars['&Lambda;'] = 'Λ'  # Lambda Λ
        chars['&Mu;'] = 'Μ'  # Mu
        chars['&Nu;'] = 'Ν'  # Nu
        chars['&Xi;'] = 'Ξ'  # Xi
        chars['&Omicron;'] = 'Ο'  # Omicron
        chars['&Pi;'] = 'Π'  # Pi Π
        chars['&Rho;'] = 'Ρ'  # Rho
        chars['&Sigma;'] = 'Σ'  # Sigma   Σ
        chars['&Tau;'] = 'Τ'  # Tau
        chars['&Upsilon;'] = 'Υ'  # Upsilon
        chars['&Phi;'] = 'Φ'  # Phi       Φ
        chars['&Chi;'] = 'Χ'  # Chi
        chars['&Psi;'] = 'Ψ'  # Psi       Ψ
        chars['&Omega;'] = 'Ω'  # Omega   Ω
        chars['&alpha;'] = 'α'  # alpha   α
        chars['&beta;'] = 'β'  # beta     β
        chars['&gamma;'] = 'γ'  # gamma   γ
        chars['&delta;'] = 'δ'  # delta   δ
        chars['&epsilon;'] = 'ε'  # epsilon       ε
        chars['&zeta;'] = 'ζ'  # zeta     ζ
        chars['&eta;'] = 'η'  # eta       η
        chars['&theta;'] = 'θ'  # theta   θ
        chars['&iota;'] = 'ι'  # iota     ι
        chars['&kappa;'] = 'κ'  # kappa   κ
        chars['&lambda;'] = 'λ'  # lambda λ
        chars['&mu;'] = 'μ'  # mu μ
        chars['&nu;'] = 'ν'  # nu ν
        chars['&xi;'] = 'ξ'  # xi ξ
        chars['&omicron;'] = 'ο'  # omicron       ο
        chars['&pi;'] = 'π'  # pi π
        chars['&rho;'] = 'ρ'  # rho       ρ
        chars['&sigmaf;'] = 'ς'  # sigmaf ς
        chars['&sigma;'] = 'σ'  # sigma   σ
        chars['&tau;'] = 'τ'  # tau       τ
        chars['&upsilon;'] = 'υ'  # upsilon       υ
        chars['&phi;'] = 'φ'  # phi       φ
        chars['&chi;'] = 'χ'  # chi       χ
        chars['&psi;'] = 'ψ'  # psi       ψ
        chars['&omega;'] = 'ω'  # omega   ω
        chars['&thetasym;'] = 'ϑ'  # theta symbol ϑ
        chars['&upsih;'] = 'ϒ'  # upsilon symbol  ϒ
        chars['&piv;'] = 'ϖ'  # pi symbol ϖ
        chars['\u03BF'] = 'ο'  # GREEK SMALL LETTER OMICRON

    if lang.lower() not in cjk_langs:
        chars['\uFF5B'] = '{'  # ｛
        chars['\uFF5D'] = '}'  # ｝
        chars['\uFF08'] = '('  # （
        chars['\uFF09'] = ')'  # ）
        chars['\uFF3B'] = '['  # ［
        chars['\uFF3D'] = ']'  # ］
        chars['\u3010'] = '('  # 【
        chars['\u3011'] = ')'  # 】
        chars['\u3002'] = '.'  # 。
        chars['\u3001'] = ','  # 、
        chars['\uFF0C'] = ','  # ，
        chars['\uFF1A'] = ':'  # ：
        chars['\uFF1B'] = ';'  # ；
        chars['\uFF1F'] = '?'  # ？
        chars['\uFF01'] = '!'  # ！
        chars['\uFF1C'] = '<'  # ＜
        chars['\uFF1D'] = '='  # ＝
        chars['\uFF1E'] = '>'  # ＞
        chars['\uFF3F'] = '_'  # ＿
        chars['\uFF40'] = "'"  # ｀
        '''
    else:
        chars['{'] = '\uFF5B'  # ｛
        chars['}'] = '\uFF5D'  # ｝
        chars['('] = '\uFF08'  # （
        chars[')'] = '\uFF09'  # ）
        chars['['] = '\uFF3B'  # ［
        chars[']'] = '\uFF3D'  # ］
        chars['\u3010'] = '\u3010'  # 【 -  #We maintain the same char
        chars['\u3011'] = '\u3011'  # 】 -  #We maintain the same char
        chars['\u3002'] = '\u3002'  # 。 #We maintain the same char
        chars[','] = '\u3001'  # 、
        chars[','] = '\uFF0C'  # ，
        chars['\u2026'] = '\u2026'  # … #We maintain the same char
        chars['\u2025'] = '\u2025'  # ‥ #We maintain the same char
        chars[':'] = '\uFF1A'  # ：
        chars[';'] = '\uFF1B'  # ；
        chars['?'] = '\uFF1F'  # ？
        chars['!'] = '\uFF01'  # ！
        chars['<'] = '\uFF1C'  # ＜
        chars['='] = '\uFF1D'  # ＝
        chars['>'] = '\uFF1E'  # ＞
        chars['_'] = '\uFF3F'  # ＿
        chars["'"] = '\uFF40'  # ｀
        '''

    charsRe = re.compile("(\\" + '|\\'.join(chars.keys()) + ")")

    return chars, charsRe


def getNormalizedPunctReplacements(lang):
    if lang.lower() == "fr":
        replacements = {
            " ,": ",",
            " )": ")",
            " }": "}",
            " ]": "]",
            #            " \""      :       "\"",          
            " ...": "...",
            "( ": "(",
            "{ ": "{",
            "[ ": "["
        }

    else:
        replacements = {
            " .":".",
            " !": "!",
            " ?": "?",
            " :": ":",
            " ;": ";",
            " ,": ",",
            " )": ")",
            " }": "}",
            " ]": "]",
            #            " \""	:	"\"",
            " ...": "...",
            " º": "º",

            "( ": "(",
            "{ ": "{",
            "[ ": "[",
            "¿ ": "¿",
            ",,.": ".",
            "¡ ": "¡"
        }
    if lang.lower() == 'sl':
        del replacements[' ...']
        del replacements[' :']

    rep = dict((re.escape(k), v) for k, v in replacements.items())
    pattern = re.compile("|".join(rep.keys()))
    return rep, pattern


# Orthographic corrections
def getReplacements(lang):
    replacements = {}
    input_replacements = None

    if lang.lower() in ["da", "de", "en", "es", "nb", "nl", "pt", "tr"]:
        input_replacements = open(os.path.dirname(os.path.realpath(__file__)) + "/replacements/replacements." + lang.lower(), "r")

    if input_replacements is not None:
        for i in input_replacements:
            field = i.split(u"\t")
            replacements[field[0]] = field[1].rstrip("\n")

    return replacements

# Detokenization corrections
def getDetokenizations(lang):
    detoks = {}
    input_detoks = None

    if lang.lower() in ["mt"]:
        input_detoks = open(os.path.dirname(os.path.realpath(__file__)) + "/detok/detok." + lang.lower(), "r")

    if input_detoks is not None:
        for i in input_detoks:
            fields = i.rstrip('\n').split('\t')
            detoks[fields[0]] = tuple(fields[1:])

    return detoks

def replace_chars(match):
    global global_chars_lang
    char = match.group(0)
    return global_chars_lang[char]


def replace_chars3(match):
    char = match.group(0)
    return ""


def fix(text, lang, chars_rep, chars_pattern):
    global global_chars_lang
    global_chars_lang = chars_rep

    ftfy_fixed_text = ftfy.fix_text_segment(text, uncurl_quotes=False, fix_latin_ligatures=False)

    replaced_text = chars_pattern.sub(replace_chars, ftfy_fixed_text)

    return html.unescape(replaced_text)

def normalize(text, lang, punct_rep, punct_pattern):
    normalized_text = text
    if lang.lower() not in cjk_langs:
        normalized_text = chars3Re.sub(replace_chars3, normalized_text)
    normalized_text = chars3Re2.sub(replace_chars3, normalized_text)
    normalized_text = chars3Re3.sub(replace_chars3, normalized_text)
    normalized_text = quotesRegex.sub("\g<start>\'\g<end>", normalized_text)
    collapsed_spaces_with_collapsed_spaces = re.sub('\s+', ' ', normalized_text)  # Collapse multiple spaces
    normalized_text_with_normalized_punct = punct_pattern.sub(lambda m: punct_rep[re.escape(m.group(0))], collapsed_spaces_with_collapsed_spaces)

    collapsed_spaces = re.sub('\s+', ' ', normalized_text_with_normalized_punct)  # Collapse multiple spaces again
    collapsed_entities = collapse_spaced_entities.sub("&#\\2;", collapsed_spaces)

    return collapsed_entities.strip(" \n")

def preserve_case(orig, dest):
    restored = ""
    if orig[0].isupper():
        restored = dest[0].upper()
    else:
        restored = dest[0]

    if len(orig) > 1:
        if orig[1:].isupper():
            restored += dest[1:].upper()
        else:
            restored += dest[1:]

    return restored

def ortho_detok_fix(text, replacements, detoks):
    if len(replacements) > 0 or len(detoks) > 0:
        last = 0
        line = []

        # Pseudo-tokenization separate words and spaces/punct
        for j in regex.finditer(r"([^-'[:alpha:]](?:[^-[:alpha:]']*[^-'[:alpha:]])?)", text):
            if last != j.start():
                line.append((text[last:j.start()], "w"))
            line.append((text[j.start():j.end()], "s"))
            last = j.end()
        else:
            if last != len(text):
                line.append((text[last:], "w"))

        fixed_text = ""
        skip_space = 0
        for i, tok in enumerate(line):
            if tok[1] == "w":
                # Print replacement if exist, otherwise print word as is
                # preserve original case if the replacement is lowercase in both sides
                if tok[0].lower() in replacements and replacements[tok[0].lower()].islower():
                    replacement = replacements[tok[0].lower()]
                    fixed_text += preserve_case(tok[0], replacement)
                elif tok[0] in replacements:
                    fixed_text += replacements[tok[0]]
                else:
                    fixed_text += tok[0]

                # If it is a tokenization issue, mark next space to be skipped
                # the current word is in the detok list
                # the words in detok are present and separated by space
                if tok[0] in detoks \
                        and len(line) > i+2 \
                        and detoks[tok[0]][0] == line[i+2][0] \
                        and line[i+1][0] == ' ':

                    # Check if the detok requires a second space to be skipped
                    if len(line) > i+4 \
                            and len(detoks[tok[0]]) == 2 \
                            and line[i+3][0] == ' ':
                        skip_space = 2
                    else:
                        skip_space = 1

            # Print spaces if they are not marked to be skipped
            elif not skip_space:
                fixed_text += tok[0]
            else:
                skip_space -= 1

    else:
        fixed_text = text

    return fixed_text


def remove_html_tags(text):
    return re.sub(html_tags_regex, ' ', text).strip()


#!/usr/bin/env python     

__author__ = "Marta Bañón (mbanon)"
__version__ = "Version 0.1 # 21/06/2019 # Initial release # Marta Bañón"
__version__ = "Version 0.2 # 23/07/2019 # Non-Redis Bifixer # Marta Bañón"
__version__ = "Version 0.3 # 20/08/2019 # New feature: Segmentation # Marta Bañón"
__version__ = "Version 0.4 # 03/02/2020 # Monofixer # Marta Bañón"


import os
import sys
import argparse
import copy
import traceback
import logging 

from unicodedata import category as cat
from unidecode import unidecode
from tempfile import gettempdir
from timeit import default_timer
from xxhash import xxh64

try:
    from . import util
    from . import restorative_cleaning
    from . import segmenter
except (ImportError, SystemError):
    import  util    
    import restorative_cleaning
    import segmenter

# Translate table to remove non alphabetic characters
tbl = [chr(i) for i in range(sys.maxunicode) if not cat(chr(i)).startswith('L')]
remove_non_alpha = str.maketrans('', '', ''.join(tbl))


def initialization():
    global ilines
    global olines

    
    ilines = 0    
    olines = 0
    header = "--header" in sys.argv
    
    logging.info("Processing arguments...")
    parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]), formatter_class=argparse.ArgumentDefaultsHelpFormatter, description=__doc__)
    
    # Mandatory parameters
    #Input file
    parser.add_argument('input', type=argparse.FileType('rt'), default=None, help="Tab-separated file to be fixed")
    #Output file (corpus)
    parser.add_argument('output', type=argparse.FileType('w'), default=sys.stdout, help="Fixed corpus")
    #Language
    parser.add_argument("lang", type=str, help="Language of the input")

    # Options group
    groupO = parser.add_argument_group('Optional')
    #Format
    groupO.add_argument("--header", action='store_true', help="Input file will have header")
    groupO.add_argument("--scol", type=util.check_positive if not header else str, default=2 if not header else "src_text", help ="Sentence column (starting in 1). The name of the field is expected instead of the position if --header is set")
    groupO.add_argument("--sdeferredcol", type=util.check_positive if not header else str, help="Source deferred standoff annotation column (starting in 1). The name of the field is expected instead of the position if --header is set")
    groupO.add_argument("--sparagraphid", type=util.check_positive if not header else str, help="Source paragraph identification column (starting in 1). The name of the field is expected instead of the position if --header is set")

    #Character fixing
    groupO.add_argument('--ignore_characters', default=False, action='store_true', help="Doesn't fix mojibake, orthography, or other character issues")
    groupO.add_argument('--ignore_normalization', default=False, action='store_true', help="Doesn't normalize punctuation and spaces.")
    
    #Empty sides
    groupO.add_argument('--ignore_empty', default=False, action='store_true', help="Doesn't remove empty sentences")

    # Too long sentences
    groupO.add_argument('--ignore_long', default=False, action='store_true', help="Doesn't ignore too long sentences")

    #Orthography
    groupO.add_argument('--ignore_orthography', default=False, action='store_true', help="Doesn't apply orthography fixing")

    # Common detokenization issues
    groupO.add_argument('--ignore_detokenization', default=False, action='store_true', help="Doesn't fix common tokenization issues")

    
    #Deduplication
    groupO.add_argument('--ignore_duplicates', default=False, action='store_true', help="Doesn't obtain the hashes of sentences")    

    groupO.add_argument('--aggressive_dedup', default=False, action='store_true', help="Treats similar sentences as duplicates (marking them with the same hash)")

    #Segmentation
    groupO.add_argument('--ignore_segmentation' , default=False, action='store_true', help="Doesn't change segmentation of long sentences")
    groupO.add_argument('--words_before_segmenting', default=15, type=util.check_positive, help="Max words allowed in a parallel sentence before trying to segmentate it. Set to 0 to applicate segmentation on everyt33hing.")
    groupO.add_argument('--segmenter', default="nltk", type=str, choices=["nltk", "loomchild"], help="Segmenter module.")    
    groupO.add_argument('--tmp_dir', default=gettempdir(), help="Temporary directory where creating the temporary files of this program")
    
    # Logging group
    groupL = parser.add_argument_group('Logging')
    groupL.add_argument('-q', '--quiet', action='store_true', help='Silent logging mode')
    groupL.add_argument('--debug', action='store_true', help='Debug logging mode')
    groupL.add_argument('--logfile', type=argparse.FileType('a'), default=sys.stderr, help="Store log to a file")
    groupL.add_argument('-v', '--version', action='version', version="%(prog)s " + __version__, help="show version of this script and exit")
 
    # Validating & parsing
    args = parser.parse_args()
    util.logging_setup(args)
    args.dedup = not args.ignore_duplicates  #more friendly usage of the ignore_duplicates flag


    logging.debug("Arguments processed: {}".format(str(args)))
 
    logging.info("Arguments processed.")
     

    return args


def fix_sentences(args):
    if ('ilines' in globals() and 'olines' in globals()):
        global ilines
        global olines        
    else:
        ilines = 0
        olines = 0    

    if not args.ignore_characters:
        chars_lang, charsRe_lang = restorative_cleaning.getCharsReplacements(args.lang)

    if not args.ignore_normalization:
        punctChars_lang, punctRe_lang = restorative_cleaning.getNormalizedPunctReplacements(args.lang)

    if not args.ignore_orthography:    
        replacements_lang = restorative_cleaning.getReplacements(args.lang)
      
    if not args.ignore_detokenization:
        detoks_lang = restorative_cleaning.getDetokenizations(args.lang)
        
    if not args.ignore_segmentation:
        lang_segmenter = segmenter.NaiveSegmenter(args.lang, args.segmenter)

    if args.header:
        header = next(args.input).strip().split("\t")

        # Transform fields to idxs
        if args.scol not in header:
            raise Exception(f"The provided --scol '{args.scol}' is not in the input header")

        args.scol = int(header.index(args.scol)) + 1

        if args.sdeferredcol:
            if args.sdeferredcol not in header:
                raise Exception(f"The provided --sdeferredcol '{args.sdeferredcol}' is not in the input header")

            args.sdeferredcol = int(header.index(args.sdeferredcol)) + 1
        if args.sparagraphid:
            if args.sparagraphid not in header:
                raise Exception(f"The provided --sparagraphid '{args.sparagraphid}' is not in the input header")

            args.sparagraphid = int(header.index(args.sparagraphid)) + 1

        # Write the output header once
        args.output.write("\t".join(header))

        if args.dedup:
            args.output.write("\tbifixer_hash\tbifixer_score")

        args.output.write("\n")

    for i in args.input:
        ilines += 1
        parts = i.split("\t")

        try:
            sentence = parts[args.scol-1]

            # Check optional indexes
            if args.sdeferredcol:
                parts[args.sdeferredcol-1]
            if args.sparagraphid:
                parts[args.sparagraphid-1]

        except IndexError:
            logging.error(traceback.format_exc())
            logging.error("Wrong column index on line " + str(ilines))
            continue
            
        very_long = False

        if not args.ignore_long and (len(sentence) > 5000):
                very_long = True

        if not args.ignore_characters and not very_long:        
           fixed_sentence = restorative_cleaning.fix(sentence, args.lang, chars_lang, charsRe_lang)
        else:
            fixed_sentence = sentence

        if not args.ignore_normalization:
            fixed_sentence = restorative_cleaning.normalize(fixed_sentence, args.lang, punctChars_lang, punctRe_lang)
                
        if not args.ignore_orthography and not very_long:    
            corrected_sentence = restorative_cleaning.ortho_detok_fix(fixed_sentence, replacements_lang, detoks_lang)
        else:
            corrected_sentence = fixed_sentence
                
        if not args.ignore_segmentation and (len(fixed_sentence.split()) > args.words_before_segmenting) and not very_long:
            segments = segmenter.naive_segmenter_mono(lang_segmenter, corrected_sentence) 

        else:
            #keep original segmentation    
            segments = [corrected_sentence]

        sent_num = 0        

        for segment in segments:
            if not args.ignore_empty and len(segment) == 0:
                continue
            if args.dedup:
                if args.aggressive_dedup:
                    #normalized_sentence = unidecode.unidecode(segment.lower().replace(" ", "").translate(str.maketrans('', '', string.punctuation+string.digits)))
                    normalized_sentence = unidecode(segment.lower().translate(remove_non_alpha))
                    hash = xxh64(normalized_sentence).hexdigest()
                        
                    charsum = sum(ord(ch) for ch in segment)
                    if len(segment) == 0:
                        ranking = 0.0
                    else:
                        ranking = round(charsum/len(segment),2)

                else:
                    hash = xxh64(segment).hexdigest()
                    ranking = 1
            #if  dedupping: Add extra columnsn with hash and ranking in output file
            #Restored parts object, with the fixed segment, overwritten for each extra segment
            new_parts = copy.deepcopy(parts)
            new_parts[args.scol-1] = segment
            
            if len(segments) > 1:
                sent_num += 1
                if args.sdeferredcol:
                    if "#" in parts[args.sdeferredcol-1]:
                        # Reconstruction
                        if sent_num != int(parts[args.sdeferredcol-1].split('#')[1]):
                            continue
                    else:
                        new_parts[args.sdeferredcol-1] = parts[args.sdeferredcol-1].rstrip("\n")+"#"+str(sent_num)

                if args.sparagraphid:
                    new_parts[args.sparagraphid-1] = parts[args.sparagraphid-1].rstrip("\n")+"#"+str(sent_num)
            # sentence may be empty now because it contained only spaces or similar weird thing
            # for a sentence containing only spaces but not normalized, strip it
            if args.ignore_empty or (new_parts[args.scol-1].strip()):
                if (args.dedup):
                    #Remove the "/n" at the end of the last item
                    new_parts[-1]= str(new_parts[-1]).strip("\n")
                
                    new_parts.append(hash) #hash and ranking are added at the end           
                    new_parts.append(ranking)
                    args.output.write("\t".join(str(v) for v in new_parts)+"\n")  #Convert to strings
                    #Remove hash and ranking for next iterations of loop
                    new_parts.pop()
                    new_parts.pop()
                else:                   
                    #When no deduplicating:
                    args.output.write("\t".join(str(v) for v in new_parts)+"\n")
                olines += 1
            else:
                #empty sentence after processing
                continue    
                    
          
            
def perform_fixing(args):
    global ilines
    global olines
    
    
    time_start=default_timer()
    logging.info("Starting fixing text")    
    fix_sentences(args)
    logging.info("Text fixing finished")

    # Stats
    logging.info("Finished")
    elapsed_time = default_timer() - time_start
    logging.info("Input lines: {0} rows".format(ilines))
    logging.info("Output lines: {0} rows".format(olines))
    logging.info("Elapsed time {0:.2f} s".format(elapsed_time))
    logging.info("Troughput: {0} rows/s".format(int((ilines*1.0)/elapsed_time)))

    logging.info("Output file: {0}".format(os.path.abspath(args.output.name)))

    
def main(args):
    logging.info("Executing main program...")
    perform_fixing(args)
    logging.info("Program finished")    


if __name__ == '__main__':
    try:
        util.logging_setup()
        args = initialization() # Parsing parameters
        main(args)  # Running main program
    except Exception as ex:
        tb = traceback.format_exc()
        logging.error(tb)
        sys.exit(1)
        



#!/usr/bin/env python

import sys
import argparse
import logging


# Logging config
def logging_setup(args=None):
    logger = logging.getLogger()
    logger.handlers = []  # Removing default handler to avoid duplication of log messages
    logger.setLevel(logging.ERROR)

    h = logging.StreamHandler(sys.stderr)
    if args != None:
        h = logging.StreamHandler(args.logfile)

    h.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
    logger.addHandler(h)

    # logger.setLevel(logging.INFO)

    if args != None:
        if not args.quiet:
            logger.setLevel(logging.INFO)
        if args.debug:
            logger.setLevel(logging.DEBUG)


# Check if the argument of a program (argparse) is strictly positive
def check_positive(value):
    ivalue = int(value)
    if ivalue <= 0:
        raise argparse.ArgumentTypeError("%s is an invalid positive int value" % value)
    return ivalue



#!/usr/bin/env python     

__author__ = "Marta Bañón (mbanon)"
__version__ = "Version 0.1 # 03/07/2019 # Initial release # Marta Bañón"
__version__ = "Version 0.2 # 23/08/2019 # Included NLTK segmenter # Marta Bañón"

import signal
import json
import os
import nltk
import sys
import logging

from nltk import load

try:
    from loomchild.segmenter import LoomchildSegmenter
except ImportError:
    pass

class myTimeout(Exception):
    pass

def timeout(signum, frame):
    raise myTimeout

class NLTKSegmenter:
    def __init__(self, lang):
        try:
            nltk.data.find('tokenizers/punkt')
        except LookupError:
            for i in range(3):
                signal.signal(signal.SIGALRM, timeout)
                signal.alarm(120)
                try:
                    result = nltk.download('punkt', quiet=True)
                    signal.alarm(0)
                    break
                except myTimeout:
                    pass
            else:
                raise Exception("Unable to download 'punkt' NLTK data after 3 retries: try to download it manually or check your internet connection.")

        langname = self.getLanguageName(lang.lower())

        try:
            self.segmenter = load('tokenizers/punkt/{0}.pickle'.format(langname))
        except:
            self.segmenter = load('tokenizers/punkt/english.pickle')

    def get_segmentation(self, sentence):
        sentence_segments = json.loads(json.dumps(self.segmenter.tokenize(sentence)))
        return sentence_segments

    def getLanguageName(self, lang):
        # Returns NLTK *.pickle file for the language, if exists. If not, returns the default (english.pickle)

        if lang == "cs":
            return "czech"

        elif lang == "da":
            return "danish"

        elif lang == "de":
            return "german"

        elif lang == "el":
            return "greek"

        elif lang == "en":
            return "english"

        elif lang == "es":
            return "spanish"

        elif lang == "et":
            return "estonian"

        elif lang == "fi":
            return "finnish"

        elif lang == "fr":
            return "frech"

        elif lang == "it":
            return "italian"

        elif lang == "nb":
            return "norwegian"

        elif lang == "nl":
            return "dutch"

        elif lang == "nn":
            return "norwegian"

        elif lang == "pl":
            return "polish"

        elif lang == "pt":
            return "portuguese"

        elif lang == "ru":
            return "russian"

        elif lang == "sl":
            return "slovene"

        elif lang == "sv":
            return "swedish"

        elif lang == "tr":
            return "turkish"

        else:
            return "english"


class NaiveSegmenter:
    def __init__(self, lang, module="nltk"):
        if module == "loomchild":
            if "loomchild.segmenter" in sys.modules:
                self.segmenter = LoomchildSegmenter(lang)
            else:
                logging.warning("Loomchild segmenter unavailable, falling back to NLTK")
                self.segmenter = NLTKSegmenter(lang)
        elif module == "nltk":
            self.segmenter = NLTKSegmenter(lang)
        else:
            self.segmenter = NLTKSegmenter(lang)

    def __call__(self, sentence):
        return self.segmenter.get_segmentation(sentence)

def naive_segmenter_mono(segmenter, sentence):
    segments = segmenter(sentence)
    return segments

def naive_segmenter(source_segmenter, target_segmenter, source, target):
    source_segments = source_segmenter(source)
    target_segments = target_segmenter(target)

    if len(source_segments) == len(target_segments):
        segments = []
        for segment_pair in zip(source_segments, target_segments):
            segment = {"source_segment": segment_pair[0], "target_segment": segment_pair[1]}
            segments.append(segment)
        return segments
    else:
        return [{"source_segment": source, "target_segment": target}]



#!/usr/bin/env python

import setuptools
import subprocess
import os.path

if __name__=="__main__":
    with open("README.md", "r") as fh:
        long_description = fh.read()
    with open("requirements.txt") as rf:
        requirements = rf.read().splitlines()
    with open("optional-requirements.txt") as ef:
        extras_requirements = ef.read().splitlines()

    setuptools.setup(
        name="bifixer",
        version="0.8",
        install_requires=requirements,
        extras_require={"loomchild": extras_requirements},
        license="GNU General Public License v3.0",
        author="Prompsit Language Engineering",
        author_email="info@prompsit.com",
        description="Tool to fix bitexts and tag near-duplicates for removal",
        maintainer="Marta Bañon",
        maintainer_email="mbanon@prompsit.com",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/bitextor/bifixer",
        packages=["bifixer"],
        package_data={"bifixer": ["replacements/*", "detok/*"]},
        classifiers=[
            "Environment :: Console",
            "Intended Audience :: Science/Research",
            "Programming Language :: Python :: 3.7",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Operating System :: POSIX :: Linux",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Text Processing :: Linguistic",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Topic :: Text Processing :: Filters"
        ],
        project_urls={
            "Bifixer on GitHub": "https://github.com/bitextor/bifixer",
            "Prompsit Language Engineering": "http://www.prompsit.com",
            "Paracrawl": "https://paracrawl.eu/"
             },
        scripts=["scripts/bifixer", "scripts/monofixer"]
    )
